//
// Created by Laura Majer on 08/04/2020.
//

#include "VRPEvalOp.h"
#include "../../ECF/ECF.h"
#include "AdaptedPermutation.h"


void VRPEvalOp::registerParameters(StateP state)
{
    state->getRegistry()->registerEntry("vrp.infile", (voidP) (new std::string), ECF::STRING);
    state->getRegistry()->registerEntry("fleetSize",(voidP) new uint, ECF::UINT);

}


bool VRPEvalOp::initialize(StateP state)
{
    if(!state->getRegistry()->isModified("vrp.infile")) {
        state->getLogger()->log(1, "Error: no input file defined for VRP! (parameter 'vrp.infile'");
        return false;
    }

    voidP sptr = state->getRegistry()->getEntry("vrp.infile"); // get parameter value
    std::string filePath = *((std::string*) sptr.get()); // convert from voidP to user defined type
    voidP fsizep = state->getRegistry()->getEntry("fleetSize"); // get parameter value
    fleetSize = *((int *) fsizep.get()); // convert from voidP to user defined type



    std::ifstream iFile(filePath.c_str());
    std::string line;
    if(!iFile.is_open()) {
        state->getLogger()->log(1, "Error: can't open input file " + filePath);
        return false;
    }

    // read dimension (number of cities)
    do {
        getline(iFile,line);
    }while(line.find("DIMENSION",0) == std::string::npos);
    std::stringstream ss(line.substr(line.find(":") + 1));
    ss >> dimension;



    // parse TSP data type
    do {
        getline(iFile,line);
    }while(line.find("EDGE_WEIGHT_TYPE",0) == std::string::npos);
    std::stringstream ss_type(line.substr(line.find(":") + 1));
    string dataType;
    ss_type >> dataType;

    if(dataType == "EUC_2D") {

        do {
            getline(iFile, line);
        } while (line.find("CAPACITY", 0) == std::string::npos);
        std::stringstream ss_capacity(line.substr(line.find(":") + 1));
        ss_capacity >> capacity;


        // read coordinates, calculate euclidian distances
        do {
            getline(iFile, line);
        } while (line.find("NODE_COORD_SECTION", 0) == std::string::npos);

        coordinates.resize(dimension);
        for (int i = 0; i < dimension; i++) {
            coordinates[i].resize(2);
            int indeks;
            iFile >> indeks;
            iFile >> coordinates[i][0];
            iFile >> coordinates[i][1];
        }

        weights.resize(dimension);
        for (int i = 0; i < dimension; i++) {
            weights[i].resize(dimension);
            for (int j = 0; j < dimension; j++) {
                double diffX = coordinates[i][0] - coordinates[j][0];
                double diffY = coordinates[i][1] - coordinates[j][1];
                weights[i][j] = (int) (0.5 + sqrt(1. * diffX * diffX + diffY * diffY));
            }
        }

        do {
            getline(iFile,line);
        }while(line.find("DEMAND_SECTION",0) == std::string::npos);

        demand.resize(dimension);
        int indeks;

        for(int i = 0; i < dimension; i++) {
            iFile >> indeks >> demand[i];
        }

        int totalDimension = dimension + fleetSize -1;

        // set dimension for permutation genotype
        state->getRegistry()->modifyEntry("AdaptedPermutation.size", (voidP) new uint(totalDimension));


        // reinitialize population with updated size
        state->getPopulation()->initialize(state);

        return true;

    }


    // if all fails:
    // unrecognized TSP data type
    state->getLogger()->log(1, "VRP initializer: can't recognize VRP data in " + filePath);
    return false;
}


FitnessP VRPEvalOp::evaluate(IndividualP individual)
{
    // minimize travel distance, so use FitnessMin
    FitnessP fitness (new FitnessMin);

    // get Permutation genotype from the individual
    AdaptedPermutation::AdaptedPermutation* perm = (AdaptedPermutation::AdaptedPermutation*) individual->getGenotype().get();
    // (you can also use boost smart pointers:)
    //PermutationP perm = boost::static_pointer_cast<Permutation::Permutation> (individual->getGenotype());

    int fitnessV = 0;
    int capacityLeft;
    // genotype Permutation keeps a vector of indexes named 'variables'
    uint size = (uint) perm->variables.size();
    std::vector<int> newPerm = perm->variables;

    int i=0;
    //broji se koliko je lokacija navedeno prije prvog vozila
    while(perm->variables[i]<this->dimension) {
        i++;
    }

    int offset = 0;
    while(offset!=i) {
        newPerm.push_back(perm->variables[offset++]);
    }

    newPerm.erase(newPerm.begin(),newPerm.begin()+i);

    int dist;
    for(i = 0; i < size; i++){

        //nasli smo indeks vozila
        if(newPerm[i] >= this->dimension) {
            capacityLeft = capacity;
            //ako je vozilo zadnje u permutaciji ne vozi nigdje
            if(i == size-1) {
                break;
            } else {
                //zapocinjemo sa zbrajanjem udaljenosti
                dist = 0;
                while(true) {
                    //ako je nakon ovog polja vozilo
                    if(newPerm[i+dist+1] >= this->dimension) {
                        //ako postoji lokacija navedena nakon vozila zbroji put od nje do 0
                        if(dist!=0) {
                            fitnessV += weights[newPerm[i+dist]][0];
                        }
                        break;

                    } else {
                        ++dist;
                        if(i + dist < size) {
                            capacityLeft -= demand[newPerm[i + dist]];
                            //ako se bilo kojem vozilu prekoraci kapacitet, vrijednost permutacije je maksimalna
                            if (capacityLeft <= 0) {
                                fitness->setValue(INT_MAX);
                                return fitness;
                            }
                            if (dist == 1) {
                                fitnessV += weights[0][newPerm[i + dist]];
                            } else {
                                fitnessV += weights[newPerm[i + dist - 1]][newPerm[i + dist]];
                            }
                        } else {
                            fitnessV += weights[newPerm[i+dist-1]][0];
                            break;
                        }
                    }
                }
            }
        }
    }

    fitness->setValue(fitnessV);
    return fitness;
}
