//
// Created by Laura Majer on 08/04/2020.
//

#include "VRPNewEvalOp.h"
#include "ECF/ECF.h"
#include "../../ECF/ECF.h"
#include "NewPermutation.h"


void VRPNewEvalOp::registerParameters(StateP state)
{
    state->getRegistry()->registerEntry("vrp.infile", (voidP) (new std::string), ECF::STRING);
    state->getRegistry()->registerEntry("fleetSize",(voidP) new uint, ECF::UINT);

}


bool VRPNewEvalOp::initialize(StateP state)
{
    if(!state->getRegistry()->isModified("vrp.infile")) {
        state->getLogger()->log(1, "Error: no input file defined for VRP! (parameter 'vrp.infile'");
        return false;
    }

    voidP sptr = state->getRegistry()->getEntry("vrp.infile"); // get parameter value
    std::string filePath = *((std::string*) sptr.get()); // convert from voidP to user defined type
    voidP fsizep = state->getRegistry()->getEntry("fleetSize"); // get parameter value
    fleetSize = *((int *) fsizep.get()); // convert from voidP to user defined type



    std::ifstream iFile(filePath.c_str());
    std::string line;
    if(!iFile.is_open()) {
        state->getLogger()->log(1, "Error: can't open input file " + filePath);
        return false;
    }

    // read dimension (number of cities)
    do {
        getline(iFile,line);
    }while(line.find("DIMENSION",0) == std::string::npos);
    std::stringstream ss(line.substr(line.find(":") + 1));
    ss >> dimension;



    // parse TSP data type
    do {
        getline(iFile,line);
    }while(line.find("EDGE_WEIGHT_TYPE",0) == std::string::npos);
    std::stringstream ss_type(line.substr(line.find(":") + 1));
    string dataType;
    ss_type >> dataType;

    if(dataType == "EUC_2D") {

        do {
            getline(iFile, line);
        } while (line.find("CAPACITY", 0) == std::string::npos);
        std::stringstream ss_capacity(line.substr(line.find(":") + 1));
        ss_capacity >> capacity;


        // read coordinates, calculate euclidian distances
        do {
            getline(iFile, line);
        } while (line.find("NODE_COORD_SECTION", 0) == std::string::npos);

        coordinates.resize(dimension);
        for (int i = 0; i < dimension; i++) {
            coordinates[i].resize(2);
            int indeks;
            iFile >> indeks;
            iFile >> coordinates[i][0];
            iFile >> coordinates[i][1];
        }

        weights.resize(dimension);
        for (int i = 0; i < dimension; i++) {
            weights[i].resize(dimension);
            for (int j = 0; j < dimension; j++) {
                double diffX = coordinates[i][0] - coordinates[j][0];
                double diffY = coordinates[i][1] - coordinates[j][1];
                weights[i][j] = (int) (0.5 + sqrt(1. * diffX * diffX + diffY * diffY));
            }
        }

        do {
            getline(iFile,line);
        }while(line.find("DEMAND_SECTION",0) == std::string::npos);

        demand.resize(dimension);
        int indeks;

        for(int i = 0; i < dimension; i++) {
            iFile >> indeks >> demand[i];
        }


        // set dimension for permutation genotype
        state->getRegistry()->modifyEntry("NewPermutation.size", (voidP) new uint(dimension-1));


        // reinitialize population with updated size
        state->getPopulation()->initialize(state);

        return true;

    }


    // if all fails:
    // unrecognized TSP data type
    state->getLogger()->log(1, "VRP initializer: can't recognize VRP data in " + filePath);
    return false;
}


FitnessP VRPNewEvalOp::evaluate(IndividualP individual)
{
    // minimize travel distance, so use FitnessMin
    FitnessP fitness (new FitnessMin);

    // get Permutation genotype from the individual
    NewPermutation::NewPermutation* perm = (NewPermutation::NewPermutation*) individual->getGenotype().get();
    // (you can also use boost smart pointers:)
    //NewPermutationP perm = boost::static_pointer_cast<NewPermutation> (individual->getGenotype());

    int fitnessV = 0;
    int capacityLeft = capacity;
    int currentWeight = 0;
    int vehiclesLeft = fleetSize - 1;
    int vCount = 1;

    // genotype Permutation keeps a vector of indexes named 'variables'
    uint size = (uint) perm->variables.size();
    std::string schedule;
    schedule += "\nVozilo 1.\t0\t";
    vCount++;

    std::vector<int> newPerm;
    newPerm.resize(size);


    for(int i=0; i<size; i++) {
        newPerm[size-i-1] = perm->variables[i];

        if(demand[perm->variables[i]] <= capacityLeft) {
            if(capacityLeft == capacity) {
                currentWeight += weights[0][perm->variables[i]];

            } else {
                currentWeight += weights[perm->variables[i-1]][perm->variables[i]];
            }
            capacityLeft -= demand[perm->variables[i]];
            schedule = schedule + std::to_string(perm->variables[i]);
            schedule = schedule + "\t";
        } else {
            if(vehiclesLeft > 0) {
                currentWeight += weights[perm->variables[i-1]][0];
                schedule += "0";
                schedule = schedule + "\n";
                fitnessV += currentWeight;
                currentWeight = weights[0][perm->variables[i]];
                schedule += "Vozilo ";
                schedule = schedule + std::to_string(vCount++);
                schedule = schedule + ".\t0\t";
                schedule = schedule + std::to_string(perm->variables[i]);
                schedule = schedule + "\t";
                capacityLeft = capacity - demand[perm->variables[i]];
                vehiclesLeft--;
            } else {
                fitness->setValue(INT_MAX);
                return fitness;
            }
        }

    }
    fitnessV += currentWeight;
    fitnessV += weights[perm->variables[size-1]][0];
    schedule += "0";

    if(perm->getDoubleCheck() == 0) {
        fitness->setValue(fitnessV);
        perm->setSchedule(schedule);
        return fitness;
    } else {


        int fitnessVnew = 0;
        capacityLeft = capacity;
        currentWeight = 0;
        vehiclesLeft = fleetSize - 1;
        vCount = 1;
        std::string new_schedule;
        new_schedule += "\nNEW";
        new_schedule += "\nVozilo 1.\t0\t";
        vCount++;


        for (int i = 0; i < size; i++) {

            if (demand[newPerm[i]] <= capacityLeft) {
                if (capacityLeft == capacity) {
                    currentWeight += weights[0][newPerm[i]];

                } else {
                    currentWeight += weights[newPerm[i - 1]][newPerm[i]];
                }
                capacityLeft -= demand[newPerm[i]];
                new_schedule = new_schedule + std::to_string(perm->variables[i]);
                new_schedule = new_schedule + "\t";
            } else {
                if (vehiclesLeft > 0) {
                    currentWeight += weights[newPerm[i - 1]][0];
                    new_schedule += "0";
                    new_schedule = new_schedule + "\n";
                    fitnessVnew += currentWeight;
                    currentWeight = weights[0][newPerm[i]];
                    new_schedule += "Vozilo ";
                    new_schedule = new_schedule + std::to_string(vCount);
                    vCount++;
                    new_schedule = new_schedule + ".\t0\t";
                    new_schedule = new_schedule + std::to_string(newPerm[i]);
                    new_schedule = new_schedule + "\t";
                    capacityLeft = capacity - demand[newPerm[i]];
                    vehiclesLeft--;
                } else {
                    fitness->setValue(INT_MAX);
                    return fitness;
                }
            }

        }
        fitnessVnew += currentWeight;
        fitnessVnew += weights[newPerm[size - 1]][0];
        //schedule += "0";

        if (fitnessVnew < fitnessV) {
            fitness->setValue(fitnessVnew);
            perm->setSchedule(new_schedule);
        } else {
            fitness->setValue(fitnessV);
            perm->setSchedule(schedule);
        }

        /*fitness->setValue(fitnessV);
        perm->setSchedule(schedule);*/
        return fitness;
    }
}
